/*
 * lh_lock.h
 *
 * Copyright (c) 1996 by Jim Goodman
 *
 * ALL RIGHTS RESERVED.
 *
 * This source file is part of the SCI protocol module for the Wisconsin
 * Wind Tunnel, written by Alain K\"agi.
 *
 * ----------------------------------------------------------------------------
 *
 * LH lock.
 */

#ifndef _lh_lock_h
#define _lh_lock_h

#include "assert.h"

/*
 * Declare a lock.
 * This pointer can be in private or shared memory as long as only one node
 * does the LH_LOCKINIT(lockname) and all nodes get the value from that.
 */
#define LH_LOCKDEC(lockname)		LHLockPtr lockname;

/*
 * Initialize a lock.
 */
#define LH_LOCKINIT(lockname)						      \
do {									      \
	int __i__;							      \
									      \
	(lockname) = (LHLockPtr)shm_malloc(sizeof(LH_master_t) +	      \
					 wwt_num_nodes * sizeof(LH_qnode_t)); \
	assert((lockname) != 0);					      \
									      \
	(lockname)->L_flag = 0;						      \
	(lockname)->L = &(lockname)->L_flag;				      \
	for (__i__ = 0; __i__ < wwt_num_nodes; __i__++) {		      \
		(lockname)->local_qnodes[__i__].I =			      \
					&(lockname)->local_qnodes[__i__].flag;\
	}								      \
} while (0);

/*
 * Calls.
 */
#define LH_LOCK(lockname)		lh_lock(lockname)
#define LH_UNLOCK(lockname)		lh_unlock(lockname)

#define LH_LOCK_ENTRY(lockname)		LH_LOCK(lockname)
#define LH_UNLOCK_ENTRY(lockname)	LH_UNLOCK(lockname)

#define LH_PREFETCH(lockname)
#define LH_UNPREFETCH(lockname)



/*
 * Details follow.
 */
#define LHDEBUG 0
#define LHPRINTDEBUG if (LHDEBUG) printf

/*
 * "Queueing info", one per processor per lock.
 */
typedef struct
{
  volatile int *I;
  volatile int *P;
  int padding1[14];
  int flag;
  int padding2[15];			/* false sharing prevention */
} LH_qnode_t;

/*
 * The main structure, one per lock.
 */
typedef struct
{
  int *L;
  int padding1[15];			/* false sharing prevention */
  int L_flag;
  int padding2[15];			/* false sharing prevention */
  LH_qnode_t local_qnodes[0];		/* one per proc. per lock, gcc hack */
} LH_master_t;

typedef LH_master_t *LHLockPtr;

INLINE void
lh_lock(LHLockPtr lock)
{
  LH_qnode_t *q;

  LHPRINTDEBUG("pe%d:   lock (%#x) enter L=%#x, *L=%d\n",
	       wwt_node_number, lock, lock->L, *lock->L);

  q = &lock->local_qnodes[wwt_node_number];

  LHPRINTDEBUG("pe%d:   lock (%#x) q=%#x\n", wwt_node_number, lock, q);

  *q->I = 1;

  LHPRINTDEBUG("pe%d:   lock (%#x) wrote 1 in %#x\n",
	       wwt_node_number, lock, q->I);

  q->P = (int *)wwt_swap_acq_rel(&lock->L, q->I);

  LHPRINTDEBUG("pe%d:   lock (%#x) spin on addr %#x\n",
	       wwt_node_number, lock, q->P);

  while (_signal_read(q->P) != 0) { ; }		/* spin */

  LHPRINTDEBUG("pe%d:   lock (%#x) done\n", wwt_node_number, lock);
}

INLINE void
lh_unlock(LHLockPtr lock)
{
  LH_qnode_t *q;

  LHPRINTDEBUG("pe%d: unlock (%#x) enter\n", wwt_node_number, lock);

  q = &lock->local_qnodes[wwt_node_number];

  LHPRINTDEBUG("pe%d: unlock (%#x) clear %#x\n", wwt_node_number, lock, q->I);

  _signal_write(q->I, 0);
  q->I = q->P;

  LHPRINTDEBUG("pe%d: unlock (%#x) new flag address %#x\n",
	       wwt_node_number, lock, q->I);
  LHPRINTDEBUG("pe%d: unlock (%#x) done\n", wwt_node_number, lock);
}

#endif _lh_lock_h
