/*
 * m_lock.h
 *
 * Copyright (c) 1996 by Jim Goodman
 *
 * ALL RIGHTS RESERVED.
 *
 * This source file is part of the SCI protocol module for the Wisconsin
 * Wind Tunnel, written by Alain K\"agi.
 *
 * ----------------------------------------------------------------------------
 *
 * M lock.
 */

#ifndef _m_lock_h
#define _m_lock_h

#include "assert.h"

/*
 * Declare a lock.
 * This pointer can be in private or shared memory as long as only one node
 * does the M_LOCKINIT(lockname) and all nodes get the value from that.
 */
#define M_LOCKDEC(lockname)  		MLockPtr lockname;

/*
 * Initialize a lock.
 */
#define M_LOCKINIT(lockname)						      \
do {									      \
	int __i__;							      \
									      \
	(lockname) = (MLockPtr)shm_malloc(sizeof(M_master_t)		      \
					+ wwt_num_nodes * sizeof(M_qnode_t)); \
	assert((lockname) != 0);					      \
									      \
	(lockname)->L_flag = 0;						      \
	(lockname)->info.as_ints.L = &(lockname)->L_flag;		      \
	(lockname)->info.as_ints.id = 0;				      \
	for (__i__ = 0; __i__ < wwt_num_nodes; __i__++) {		      \
		(lockname)->local_qnodes[__i__].I =			      \
					&(lockname)->local_qnodes[__i__].flag;\
	}								      \
} while (0);

/*
 * Calls.
 */
#define M_LOCK(lockname)		m_lock(lockname)
#define M_UNLOCK(lockname)		m_unlock(lockname)

#define M_LOCK_ENTRY(lockname)		M_LOCK(lockname)
#define M_UNLOCK_ENTRY(lockname)	M_UNLOCK(lockname)

#define M_PREFETCH(lockname)
#define M_UNPREFETCH(lockname)



/*
 * Details follow.
 */
#define MDEBUG 0
#define MPRINTDEBUG if (MDEBUG) printf

/*
 * "Queueing info", one per processor per lock.
 */
typedef struct
{
  union {
    struct {
      volatile int *P;
      int old_id;
    } as_ints;
    double as_double;
  } info;
  volatile int *I;
  int padding1[13];			/* false sharing prevention */
  int flag;
  int padding3[15];			/* false sharing prevention */
} M_qnode_t;

/*
 * The main structure, one per lock.
 */
typedef struct
{
  union {
    struct {
      int *L;
      int id;
    } as_ints;
    double as_double;
  } info;
  double padding1[7];			/* false sharing prevention & align */
  int L_flag;
  int padding2[15];			/* false sharing prevention */
  M_qnode_t local_qnodes[0];		/* one per proc. per lock, gcc hack */
} M_master_t;

typedef M_master_t *MLockPtr;

INLINE void
m_lock(MLockPtr lock)
{
  M_qnode_t *q;
  union {
    struct {
      volatile int *I;
      int id;
    } as_ints;
    double as_double;
  } var;
  double wwt_swap_acq_rel_d(double *, double);

  MPRINTDEBUG("pe%d:   lock (%#x) enter L=%#x, *L=%d\n",
	      wwt_node_number, lock, lock->info.as_ints.L,
	      *lock->info.as_ints.L);

  q = &lock->local_qnodes[wwt_node_number];

  MPRINTDEBUG("pe%d:   lock (%#x) q=%#x\n", wwt_node_number, lock, q);

  *q->I = 1;

  MPRINTDEBUG("pe%d:   lock (%#x) wrote 1 in %#x\n",
	      wwt_node_number, lock, q->I);

  var.as_ints.I = q->I;
  var.as_ints.id = wwt_node_number + 1;
  q->info.as_double = wwt_swap_acq_rel_d(&lock->info.as_double, var.as_double);

  MPRINTDEBUG("pe%d:   lock (%#x) spin on addr %#x\n",
	      wwt_node_number, lock, q->info.as_ints.P);

  if (q->info.as_ints.old_id > 0) {
    while (_signal_read(q->info.as_ints.P) != 0) { ; }	/* spin */
  }

  MPRINTDEBUG("pe%d:   lock (%#x) done\n", wwt_node_number, lock);
}

INLINE void
m_unlock(MLockPtr lock)
{
  M_qnode_t *q;
  union {
    struct {
      volatile int *I;
      int id;
    } as_ints;
    double as_double;
  } new_value, expected_value;
  int wwt_compare_and_set_acq_rel_d(double *, double, double);

  MPRINTDEBUG("pe%d: unlock (%#x) enter\n", wwt_node_number, lock);

  q = &lock->local_qnodes[wwt_node_number];

  MPRINTDEBUG("pe%d: unlock (%#x) clear %#x\n", wwt_node_number, lock, q->I);

  _signal_write(q->I, 0);

  new_value.as_ints.I = q->info.as_ints.P;
  new_value.as_ints.id = q->info.as_ints.old_id;
  expected_value.as_ints.I = q->I;
  expected_value.as_ints.id = wwt_node_number + 1;

  if (!wwt_compare_and_set_acq_rel_d(&lock->info.as_double,
				     new_value.as_double,
				     expected_value.as_double)) {
    MPRINTDEBUG("pe%d: unlock (%#x) compare failed\n", wwt_node_number, lock);
    q->I = q->info.as_ints.P;
  }

  MPRINTDEBUG("pe%d: unlock (%#x) new flag address %#x\n",
	      wwt_node_number, lock, q->I);
  MPRINTDEBUG("pe%d: unlock (%#x) done\n", wwt_node_number, lock);
}

#endif _m_lock_h
