/*
 * mcs_lock.h
 *
 * Copyright (c) 1995 by Jim Goodman
 *
 * ALL RIGHTS RESERVED.
 *
 * This source file is part of the SCI protocol module for the Wisconsin
 * Wind Tunnel, written by Alain K\"agi.
 *
 * ----------------------------------------------------------------------------
 *
 * Mellor-Crummey & Scott (MCS) lock.
 */

#ifndef _mcs_lock_h
#define _mcs_lock_h

#include "assert.h"

/*
 * Declare a lock.
 * This pointer can be in private or shared memory as long as only one node
 * does the MCS_LOCKINIT(lockname) and all nodes get the value from that.
 */
#define MCS_LOCKDEC(lockname)		MCS_master_t lockname

/*
 * Initialize a lock.
 */
#define MCS_LOCKINIT(lockname)						      \
do {									      \
	(lockname).local_qnodes = (MCS_qnode_t *)shm_malloc(		      \
					 wwt_num_nodes * sizeof(MCS_qnode_t));\
	assert((lockname).local_qnodes != 0);				      \
									      \
	(lockname).master = 0;						      \
} while (0);

/*
 * Calls.
 */
#define MCS_LOCK(lockname)		mcs_lock(&(lockname))
#define MCS_UNLOCK(lockname)		mcs_unlock(&(lockname))

#define MCS_LOCK_ENTRY(lockname)	MCS_LOCK(lockname)
#define MCS_UNLOCK_ENTRY(lockname)	MCS_UNLOCK(lockname)

#define MCS_PREFETCH(lockname)
#define MCS_UNPREFETCH(lockname)



/*
 * Details follow.
 */
#define MCSDEBUG 0
#define MCSPRINTDEBUG if (MCSDEBUG) printf

/*
 * "Queueing info", one per processor per lock.
 */
typedef struct MCS_qnode
{
  struct MCS_qnode *next;	/* who gets in when this node unlocks it */
  int locked;			/* 1 means another node has it locked */
  int pad[14];			/* false sharing prevention */
} MCS_qnode_t;

/*
 * The main structure, one per lock.
 */
typedef struct
{
  MCS_qnode_t *local_qnodes;	/* one per proc. per lock */
  int pad[15];			/* false sharing prevention */
  MCS_qnode_t *master;		/* the actual master lock */
} MCS_master_t;

typedef MCS_master_t *MCSLockPtr;


INLINE void mcs_lock(MCSLockPtr lockp)
{
  MCS_qnode_t *llock;
  MCS_qnode_t **master_lock;
  MCS_qnode_t *pred;

  asm volatile ("");

  master_lock = &lockp->master;
#ifndef NDEBUG
  MCSPRINTDEBUG("lock(0x%x) called,master:&=0x%x\n", lockp, master_lock);
#endif  NDEBUG
  llock = &lockp->local_qnodes[wwt_node_number];

#ifndef NDEBUG
  MCSPRINTDEBUG("lock(0x%x) called,master:&=0x%x(0x%x),llock:&=0x%x(0x%x)\n",
    lockp, master_lock, *(int *)master_lock, llock, *(int *)llock);
#endif  NDEBUG

  _signal_write(&llock->locked, 0);
  _signal_write(&llock->next, 0);
  pred = (MCS_qnode_t *)wwt_swap_acq_rel(master_lock, llock);
#ifndef NDEBUG
  MCSPRINTDEBUG("lock(0x%x) master:&=0x%x(0x%x),llock:&=0x%x(0x%x),pred=0x%x\n",
    lockp, master_lock, *(int *)master_lock, llock, *(int *)llock, pred);
#endif  NDEBUG
  if (pred)
  {
    _signal_write(&llock->locked, 1);
    _signal_write(&pred->next, llock);

#ifndef NDEBUG
    MCSPRINTDEBUG("lock: waiting for pred to release lock...\n");
    MCSPRINTDEBUG("lock(0x%x) master:&=0x%x(0x%x),llock:&=0x%x(0x%x)\n",
      lockp, master_lock, *(int *)master_lock, llock, *(int *)llock);
#endif  NDEBUG
    while (_signal_read(&llock->locked)) { ; }		/* Busy wait */
#ifndef NDEBUG
    MCSPRINTDEBUG("lock: pred has released lock...\n");
#endif  NDEBUG
  }

#ifndef NDEBUG
  MCSPRINTDEBUG("lock(0x%x) done,master:&=0x%x(0x%x),llock:&=0x%x(0x%x)\n",
    lockp, master_lock, *(int *)master_lock, llock, *(int *)llock);
#endif  NDEBUG

  asm volatile ("");
}

INLINE void mcs_unlock(MCSLockPtr lockp)
{
  MCS_qnode_t *llock;
  MCS_qnode_t **master_lock;
  MCS_qnode_t *old_tail, *usurper, *next;

  asm volatile ("");

  llock = &lockp->local_qnodes[wwt_node_number];

  master_lock = &lockp->master;

#ifndef NDEBUG
  MCSPRINTDEBUG("unlock(0x%x) called,master:&=0x%x(0x%x),llock:&=0x%x(0x%x)\n",
    lockp, master_lock, *(int *)master_lock, llock, *(int *)llock);
#endif  NDEBUG
  if ((next = (MCS_qnode_t *)_signal_read(&llock->next)) == 0)
  {
    /* No known Successor */
#ifndef NDEBUG
    MCSPRINTDEBUG("unlock: next=0x%x\n", next);
#endif  NDEBUG
    old_tail = (MCS_qnode_t *)wwt_swap_acq_rel(master_lock, 0);
#ifndef NDEBUG
    MCSPRINTDEBUG("unlock: old_tail = 0x%x\n", old_tail);
#endif  NDEBUG
    if (old_tail == llock)
    {
      /* There really was no successor */
#ifndef NDEBUG
      MCSPRINTDEBUG("unlock done, (old_tail=llock=0x%x\n", old_tail);
#endif  NDEBUG
      return;
    }
    /*
     * We have accidentally removed some nodes from the queue;
     * we need to put them back.
     */
    usurper = (MCS_qnode_t *)wwt_swap_acq_rel(master_lock, old_tail);
#ifndef NDEBUG
    MCSPRINTDEBUG("unlock: old_tail=0x%x,usurper=0x%x,waiting...\n",
      old_tail, usurper);
#endif  NDEBUG
    while ((next = (MCS_qnode_t *)_signal_read(&llock->next)) == 0)
    {
      ;  /* Wait for next node to be attached */
    }
    if (usurper)
    {
      /*
       * Somebody else got into the queue
       * Link our victims after usurper(s)
       */
      _signal_write(&usurper->next, next);
    }
    else
    {
       _signal_write(&next->locked, 0);
    }
  }
  else
  {
#ifndef NDEBUG
    MCSPRINTDEBUG("unlock: next=0x%x,&next->locked=0x%x\n",
		  next, &next->locked);
#endif  NDEBUG
    _signal_write(&next->locked, 0);
  }
#ifndef NDEBUG
  MCSPRINTDEBUG("unlock(0x%x) done,master:&=0x%x(0x%x),llock:0x%x\n",
    lockp, master_lock, *(int *)master_lock, llock);
#endif  NDEBUG

  asm volatile ("");
}

#endif /* _mcs_lock_h */
