/*
 * reactive2_lock.h
 *
 * Copyright (c) 1996 by Jim Goodman
 *
 * ALL RIGHTS RESERVED.
 *
 * This source file is part of the SCI protocol module for the Wisconsin
 * Wind Tunnel, written by Alain K\"agi.
 *
 * ----------------------------------------------------------------------------
 *
 * Reactive synchronization: Test&Set w/ backoff + MCS.
 */

#ifndef _reactive2_lock_h
#define _reactive2_lock_h

#include "assert.h"

/*
 * Declare a lock.
 * This pointer can be in private or shared memory as long as only one node
 * does the REACTIVE2_LOCKINIT(lockname) and all nodes get the value from that.
 */
#define REACTIVE2_LOCKDEC(lockname)		Reactive2LockPtr lockname

/*
 * Initialize a lock.
 */
#define REACTIVE2_LOCKINIT(lockname)					      \
do {									      \
	int __i__;							      \
									      \
	(lockname) = (Reactive2LockPtr)shm_malloc(sizeof(Reactive2_master_t) +\
			     (wwt_num_nodes + 1) * sizeof(Reactive2_qnode_t));\
	assert((lockname) != 0);					      \
									      \
	(lockname)->delays = (Reactive2_delay_t *)shm_malloc(wwt_num_nodes    \
						* sizeof(Reactive2_delay_t)); \
									      \
	for (__i__ = 0; __i__ < wwt_num_nodes; __i__++) {		      \
		(lockname)->delays[__i__].delay = 2;			      \
	}								      \
									      \
	(lockname)->mode = REACTIVE2_TSBACK;				      \
	(lockname)->lock = 0;		/* TSBACK lock is free */	      \
					/* QUEUE lock is busy */	      \
	(lockname)->master = &(lockname)->local_qnodes[wwt_num_nodes];	      \
	(lockname)->local_qnodes[wwt_num_nodes].locked = 1;		      \
	(lockname)->local_qnodes[wwt_num_nodes].next = 0;		      \
	(lockname)->q_empty_count = 0;					      \
	(lockname)->high_delay_count = 0;				      \
} while (0);

/*
 * Calls.
 */
#define REACTIVE2_LOCK(lockname)		reactive2_lock(lockname)
#define REACTIVE2_UNLOCK(lockname)		reactive2_unlock(lockname)

#define REACTIVE2_LOCK_ENTRY(lockname)		REACTIVE2_LOCK(lockname)
#define REACTIVE2_UNLOCK_ENTRY(lockname)	REACTIVE2_UNLOCK(lockname)

#define REACTIVE2_PREFETCH(lockname)
#define REACTIVE2_UNPREFETCH(lockname)



/*
 * Details follow.
 */

#define REACTIVE2_MIN_DELAY		(2)
#define REACTIVE2_MAX_DELAY		(2 * 512 * wwt_num_nodes)
#define REACTIVE2_THRESHOLD_DELAY	(64)
#define REACTIVE2_MIN(A,B)		(((A) < (B)) ? (A) : (B))
#define REACTIVE2_MAX(A,B)		(((A) > (B)) ? (A) : (B))

#define REACTIVE2DEBUG 0
#define REACTIVE2PRINTDEBUG if (REACTIVE2DEBUG) printf

/*
 * "Queueing info", one per processor per lock.
 */
typedef struct Reactive2_qnode
{
  struct Reactive2_qnode *next;	/* who gets in when this node unlocks it */
  int locked;			/* 1 means another node has it locked */
  int pad[14];			/* false sharing prevention */
} Reactive2_qnode_t;

typedef enum { REACTIVE2_TSBACK, REACTIVE2_QUEUE } Reactive2_policy_t;

typedef struct
{
  int delay;
  int pad[15];
} Reactive2_delay_t;

/*
 * The main structure, one per lock.
 */
typedef struct
{
  Reactive2_delay_t *delays;
  int delaypad[15];
  volatile Reactive2_policy_t mode;
  int modepad[15];
  int lock;				/* used when mode == TS */
  Reactive2_qnode_t *master;		/* used when mode == QUEUE */
  unsigned int q_empty_count;
  unsigned int high_delay_count;
  int pad[12];				/* false sharing prevention */
  Reactive2_qnode_t local_qnodes[0];	/* one per proc. per lock, gcc hack */
} Reactive2_master_t;

typedef Reactive2_master_t *Reactive2LockPtr;

INLINE void
reactive2_lock(Reactive2LockPtr lockp)
{
  Reactive2_qnode_t *llock;
  Reactive2_qnode_t **master_lock;
  Reactive2_qnode_t *pred;
  int *lock;
  int delay;

 retry:
  switch (lockp->mode)
  {
  case REACTIVE2_TSBACK:
    lock = &lockp->lock;

    if (wwt_swap_acq(lock, 1))
    {
      delay = lockp->delays[wwt_node_number].delay;

      delay = REACTIVE2_MAX(REACTIVE2_MIN_DELAY, delay/2);
      wwt_delay(random() % delay);

      while (wwt_swap_acq(lock, 1)
	     && lockp->mode == REACTIVE2_TSBACK)
      {
	delay = REACTIVE2_MIN(REACTIVE2_MAX_DELAY, delay*2);
	wwt_delay(random() % delay);
      }

      lockp->delays[wwt_node_number].delay = delay;
    }

    if (lockp->mode == REACTIVE2_QUEUE)
    {
      goto retry;
    }

    if (delay > REACTIVE2_THRESHOLD_DELAY)
    {
      if (lockp->high_delay_count++ > 5)
      {
	REACTIVE2PRINTDEBUG("%d: switching to QUEUE, delay = %d\n",
			    wwt_node_number, delay);
	lockp->high_delay_count = 0;
	llock = &lockp->local_qnodes[wwt_node_number];
	llock->locked = 0;
	llock->next = 0;
	lockp->master = llock;
	lockp->mode = REACTIVE2_QUEUE;
      }
    }
    else
    {
      lockp->high_delay_count = 0;
    }

    break;

  case REACTIVE2_QUEUE:
    master_lock = &lockp->master;
    REACTIVE2PRINTDEBUG("lock(0x%x) called,master:&=0x%x\n",
		       lockp, master_lock);
    llock = &lockp->local_qnodes[wwt_node_number];

    REACTIVE2PRINTDEBUG("lock(0x%x) called,master:&=0x%x(0x%x),llock:&=0x%x(0x%x)\n",
		       lockp, master_lock, *(int *)master_lock, llock,
		       *(int *)llock);

    llock->locked = 0;
    llock->next = 0;
    pred = (Reactive2_qnode_t *)wwt_swap_acq_rel(master_lock, llock);
    REACTIVE2PRINTDEBUG("lock(0x%x) master:&=0x%x(0x%x),llock:&=0x%x(0x%x),pred=0x%x\n",
		       lockp, master_lock, *(int *)master_lock, llock,
		       *(int *)llock, pred);
    if (pred)
    {
      lockp->q_empty_count = 0;
      llock->locked = 1;
      _signal_write(&pred->next, llock);

      REACTIVE2PRINTDEBUG("lock: waiting for pred to release lock...\n");
      REACTIVE2PRINTDEBUG("lock(0x%x) master:&=0x%x(0x%x),llock:&=0x%x(0x%x)\n",
			 lockp, master_lock, *(int *)master_lock, llock,
			 *(int *)llock);
      while (_signal_read(&llock->locked) && lockp->mode == REACTIVE2_QUEUE)
      {
	;		/* Busy wait */
      }

      if (lockp->mode == REACTIVE2_TSBACK)
      {
	goto retry;
      }

      REACTIVE2PRINTDEBUG("lock: pred has released lock...\n");
    }
    else
    {
      if (lockp->q_empty_count++ > 4)
      {
	REACTIVE2PRINTDEBUG("%d: switching to TSBACK\n", wwt_node_number);
	lockp->lock = 1;
	lockp->q_empty_count = 0;
	lockp->mode = REACTIVE2_TSBACK;
      }
    }

    REACTIVE2PRINTDEBUG("lock(0x%x) done,master:&=0x%x(0x%x),llock:&=0x%x(0x%x)\n",
		       lockp, master_lock, *(int *)master_lock, llock,
		       *(int *)llock);
    break;

  }

  asm volatile ("");
}

INLINE void
reactive2_unlock(Reactive2LockPtr lockp)
{
  Reactive2_qnode_t *llock;
  Reactive2_qnode_t **master_lock;
  Reactive2_qnode_t *old_tail, *usurper, *next;

  asm volatile ("");

  switch (lockp->mode)
  {
  case REACTIVE2_TSBACK:
    _signal_write(&lockp->lock, 0);

    break;

  case REACTIVE2_QUEUE:
    llock = &lockp->local_qnodes[wwt_node_number];

    master_lock = &lockp->master;

    REACTIVE2PRINTDEBUG("unlock(0x%x) called,master:&=0x%x(0x%x),llock:&=0x%x(0x%x)\n",
		       lockp, master_lock, *(int *)master_lock, llock,
		       *(int *)llock);
    if ((next = (Reactive2_qnode_t *)_signal_read(&llock->next)) == 0)
    {
      /* No known Successor */
      REACTIVE2PRINTDEBUG("unlock: next=0x%x\n", next);
      old_tail = (Reactive2_qnode_t *)wwt_swap_acq_rel(master_lock, 0);
      REACTIVE2PRINTDEBUG("unlock: old_tail = 0x%x\n", old_tail);
      if (old_tail == llock)
      {
	/* There really was no successor */
	lockp->q_empty_count++;
	REACTIVE2PRINTDEBUG("unlock done, (old_tail=llock=0x%x\n", old_tail);
	return;
      }
      /*
       * We have accidentally removed some nodes from the queue;
       * we need to put them back.
       */
      usurper = (Reactive2_qnode_t *)wwt_swap_acq_rel(master_lock, old_tail);
      REACTIVE2PRINTDEBUG("unlock: old_tail=0x%x,usurper=0x%x,waiting...\n",
			 old_tail, usurper);
      while ((next = (Reactive2_qnode_t *)_signal_read(&llock->next)) == 0)
      {
	;  /* Wait for next node to be attached */
      }
      if (usurper)
      {
	/*
	 * Somebody else got into the queue
	 * Link our victims after usurper(s)
	 */
	_signal_write(&usurper->next, next);
      }
      else
      {
	_signal_write(&next->locked, 0);
      }
    }
    else
    {
      REACTIVE2PRINTDEBUG("unlock: next=0x%x,&next->locked=0x%x\n",
			 next, &next->locked);
      _signal_write(&next->locked, 0);
    }
    REACTIVE2PRINTDEBUG("unlock(0x%x) done,master:&=0x%x(0x%x),llock:0x%x\n",
		       lockp, master_lock, *(int *)master_lock, llock);

    break;
  }
}

#endif _reactive2_lock_h
