/*
 * reactive_lock.h
 *
 * Copyright (c) 1996 by Jim Goodman
 *
 * ALL RIGHTS RESERVED.
 *
 * This source file is part of the SCI protocol module for the Wisconsin
 * Wind Tunnel, written by Alain K\"agi.
 *
 * ----------------------------------------------------------------------------
 *
 * Reactive synchronization: Test&Set + MCS.
 */

#ifndef _reactive_lock_h
#define _reactive_lock_h

#include "assert.h"

/*
 * Declare a lock.
 * This pointer can be in private or shared memory as long as only one node
 * does the REACTIVE_LOCKINIT(lockname) and all nodes get the value from that.
 */
#define REACTIVE_LOCKDEC(lockname)	ReactiveLockPtr lockname

/*
 * Initialize a lock.
 */
#define REACTIVE_LOCKINIT(lockname)					      \
do {									      \
	(lockname) = (ReactiveLockPtr)shm_malloc(sizeof(Reactive_master_t) +  \
			      (wwt_num_nodes + 1) * sizeof(Reactive_qnode_t));\
	assert((lockname) != 0);					      \
									      \
	(lockname)->mode = REACTIVE_TS;					      \
	(lockname)->lock = 0;		/* TS lock is free */		      \
					/* QUEUE lock is busy */	      \
	(lockname)->master = &(lockname)->local_qnodes[wwt_num_nodes];	      \
	(lockname)->local_qnodes[wwt_num_nodes].locked = 1;		      \
	(lockname)->local_qnodes[wwt_num_nodes].next = 0;		      \
	(lockname)->q_empty_count = 0;					      \
	(lockname)->high_iter_count = 0;				      \
} while (0);

/*
 * Calls.
 */
#define REACTIVE_LOCK(lockname)		reactive_lock(lockname)
#define REACTIVE_UNLOCK(lockname)	reactive_unlock(lockname)

#define REACTIVE_LOCK_ENTRY(lockname)	REACTIVE_LOCK(lockname)
#define REACTIVE_UNLOCK_ENTRY(lockname)	REACTIVE_UNLOCK(lockname)

#define REACTIVE_PREFETCH(lockname)
#define REACTIVE_UNPREFETCH(lockname)



/*
 * Details follow.
 */

#define REACTIVEDEBUG 0
#define REACTIVEPRINTDEBUG if (REACTIVEDEBUG) printf

/*
 * "Queueing info", one per processor per lock.
 */
typedef struct Reactive_qnode
{
  struct Reactive_qnode *next;	/* who gets in when this node unlocks it */
  int locked;			/* 1 means another node has it locked */
  int pad[14];			/* false sharing prevention */
} Reactive_qnode_t;

typedef enum { REACTIVE_TS, REACTIVE_QUEUE } Reactive_policy_t;

/*
 * The main structure, one per lock.
 */
typedef struct
{
  Reactive_policy_t mode;
  int pad1[15];
  int lock;			/* used when mode == TS */
  Reactive_qnode_t *master;	/* used when mode == QUEUE */
  unsigned int q_empty_count;
  unsigned int high_iter_count;
  int pad[12];			/* false sharing prevention */
  Reactive_qnode_t local_qnodes[0];	/* one per proc. per lock, gcc hack */
} Reactive_master_t;

typedef Reactive_master_t *ReactiveLockPtr;

INLINE void
reactive_lock(ReactiveLockPtr lockp)
{
  Reactive_qnode_t *llock;
  Reactive_qnode_t **master_lock;
  Reactive_qnode_t *pred;
  int *lock;
  int count = 0;

 retry:
  switch (lockp->mode)
  {
  case REACTIVE_TS:
    lock = &lockp->lock;

    while (wwt_swap_acq(lock, 1) && lockp->mode == REACTIVE_TS)
    {
      count++;
    }

    if (lockp->mode == REACTIVE_QUEUE)
    {
      goto retry;
    }

    if (count > 100)
    {
      if (lockp->high_iter_count++ > 5)
      {
	REACTIVEPRINTDEBUG("%d: switching to QUEUE, count = %d\n",
			   wwt_node_number, count);
	lockp->high_iter_count = 0;
	llock = &lockp->local_qnodes[wwt_node_number];
	llock->locked = 0;
	llock->next = 0;
	lockp->master = llock;
	lockp->mode = REACTIVE_QUEUE;
      }
    }
    else
    {
      lockp->high_iter_count = 0;
    }

    break;

  case REACTIVE_QUEUE:
    master_lock = &lockp->master;
    REACTIVEPRINTDEBUG("lock(0x%x) called,master:&=0x%x\n",
		       lockp, master_lock);
    llock = &lockp->local_qnodes[wwt_node_number];

    REACTIVEPRINTDEBUG("lock(0x%x) called,master:&=0x%x(0x%x),llock:&=0x%x(0x%x)\n",
		       lockp, master_lock, *(int *)master_lock, llock,
		       *(int *)llock);

    llock->locked = 0;
    llock->next = 0;
    pred = (Reactive_qnode_t *)wwt_swap_acq_rel(master_lock, llock);
    REACTIVEPRINTDEBUG("lock(0x%x) master:&=0x%x(0x%x),llock:&=0x%x(0x%x),pred=0x%x\n",
		       lockp, master_lock, *(int *)master_lock, llock,
		       *(int *)llock, pred);
    if (pred)
    {
      llock->locked = 1;
      _signal_write(&pred->next, llock);

      REACTIVEPRINTDEBUG("lock: waiting for pred to release lock...\n");
      REACTIVEPRINTDEBUG("lock(0x%x) master:&=0x%x(0x%x),llock:&=0x%x(0x%x)\n",
			 lockp, master_lock, *(int *)master_lock, llock,
			 *(int *)llock);
      while (_signal_read(&llock->locked) && lockp->mode == REACTIVE_QUEUE)
      {
	;		/* Busy wait */
      }

      if (lockp->mode == REACTIVE_TS)
      {
	goto retry;
      }

      REACTIVEPRINTDEBUG("lock: pred has released lock...\n");
    }
    else
    {
      if (lockp->q_empty_count++ > 5)
      {
	REACTIVEPRINTDEBUG("%d: switching to TS\n", wwt_node_number);
	lockp->lock = 1;
	lockp->q_empty_count = 0;
	lockp->mode = REACTIVE_TS;
      }
    }

    REACTIVEPRINTDEBUG("lock(0x%x) done,master:&=0x%x(0x%x),llock:&=0x%x(0x%x)\n",
		       lockp, master_lock, *(int *)master_lock, llock,
		       *(int *)llock);
    break;

  }
}

INLINE void
reactive_unlock(ReactiveLockPtr lockp)
{
  Reactive_qnode_t *llock;
  Reactive_qnode_t **master_lock;
  Reactive_qnode_t *old_tail, *usurper, *next;

  switch (lockp->mode)
  {
  case REACTIVE_TS:
    asm volatile ("");
    lockp->lock = 0;

    break;

  case REACTIVE_QUEUE:
    llock = &lockp->local_qnodes[wwt_node_number];

    master_lock = &lockp->master;

    REACTIVEPRINTDEBUG("unlock(0x%x) called,master:&=0x%x(0x%x),llock:&=0x%x(0x%x)\n",
		       lockp, master_lock, *(int *)master_lock, llock,
		       *(int *)llock);
    if ((next = (Reactive_qnode_t *)_signal_read(&llock->next)) == 0)
    {
      /* No known Successor */
      REACTIVEPRINTDEBUG("unlock: next=0x%x\n", next);
      old_tail = (Reactive_qnode_t *)wwt_swap_acq_rel(master_lock, 0);
      REACTIVEPRINTDEBUG("unlock: old_tail = 0x%x\n", old_tail);
      if (old_tail == llock)
      {
	/* There really was no successor */
	lockp->q_empty_count++;
	REACTIVEPRINTDEBUG("unlock done, (old_tail=llock=0x%x\n", old_tail);
	return;
      }
      /*
       * We have accidentally removed some nodes from the queue;
       * we need to put them back.
       */
      usurper = (Reactive_qnode_t *)wwt_swap_acq_rel(master_lock, old_tail);
      REACTIVEPRINTDEBUG("unlock: old_tail=0x%x,usurper=0x%x,waiting...\n",
			 old_tail, usurper);
      while ((next = (Reactive_qnode_t *)_signal_read(&llock->next)) == 0)
      {
	;  /* Wait for next node to be attached */
      }
      if (usurper)
      {
	/*
	 * Somebody else got into the queue
	 * Link our victims after usurper(s)
	 */
	_signal_write(&usurper->next, next);
      }
      else
      {
	_signal_write(&next->locked, 0);
      }
    }
    else
    {
      REACTIVEPRINTDEBUG("unlock: next=0x%x,&next->locked=0x%x\n",
			 next, &next->locked);
      _signal_write(&next->locked, 0);
    }
    REACTIVEPRINTDEBUG("unlock(0x%x) done,master:&=0x%x(0x%x),llock:0x%x\n",
		       lockp, master_lock, *(int *)master_lock, llock);

    break;
  }
}

#endif _reactive_lock_h
