/*
 * tsback_lock.h
 *
 * Copyright (c) 1996 by Jim Goodman
 *
 * ALL RIGHTS RESERVED.
 *
 * This source file is part of the SCI protocol module for the Wisconsin
 * Wind Tunnel, written by Alain K\"agi.
 *
 * ----------------------------------------------------------------------------
 *
 * Test&set with exponential backoff (delay between references).
 */

#ifndef _tsback_lock_h
#define _tsback_lock_h

#include "assert.h"

/*
 * Declare a lock.
 */
#define TSBACK_LOCKDEC(lockname)	TSBACK_master_t lockname

/*
 * Initialize a lock.
 */
#define TSBACK_LOCKINIT(lockname)					      \
do {									      \
	int __i__;							      \
									      \
	(lockname).delays = (TSBACK_delay_t *)shm_malloc(wwt_num_nodes	      \
						* sizeof(TSBACK_delay_t));    \
	assert((lockname).delays != 0);					      \
									      \
	for (__i__ = 0; __i__ < wwt_num_nodes; __i__++) {		      \
		(lockname).delays[__i__].delay = TSBACK_MIN_DELAY;	      \
	}								      \
	(lockname).lock = 0;						      \
} while (0);

/*
 * Calls.
 */
#define TSBACK_LOCK(lockname)		tsback_lock(&(lockname))
#define TSBACK_UNLOCK(lockname)		tsback_unlock(&(lockname))

#define TSBACK_LOCK_ENTRY(lockname)	tsback_lock_entry(&(lockname))
#define TSBACK_UNLOCK_ENTRY(lockname)	tsback_unlock_entry(&(lockname))

#define TSBACK_PREFETCH(lockname)
#define TSBACK_UNPREFETCH(lockname)



/*
 * Details follow.
 */

#define TSBACK_MIN_DELAY		(2)
#define TSBACK_MAX_DELAY		(1024 * wwt_num_nodes)
#define TSBACK_MIN(A,B)			(((A) < (B)) ? (A) : (B))
#define TSBACK_MAX(A,B)			(((A) > (B)) ? (A) : (B))

#define TSBACKDEBUG 0
#define TSBACKPRINTDEBUG if (TSBACKDEBUG) printf

typedef struct
{
  int delay;
  int pad[15];
} TSBACK_delay_t;

/*
 * The main structure, one per lock.
 */
typedef struct
{
  TSBACK_delay_t *delays;
  int pad[15];
  int lock;
} TSBACK_master_t;

long random();

INLINE void
tsback_lock(TSBACK_master_t *masterp)
{
  int *lockp = &masterp->lock;

  if (wwt_swap_acq(lockp, 1)) {
    int delay = masterp->delays[wwt_node_number].delay;

    delay = TSBACK_MAX(TSBACK_MIN_DELAY, delay/2);
    wwt_delay(random() % delay);

    while (wwt_swap_acq(lockp, 1))
    {
      delay = TSBACK_MIN(TSBACK_MAX_DELAY, delay*2);
      wwt_delay(random() % delay);
    }

    masterp->delays[wwt_node_number].delay = delay;
  }
}

INLINE void
tsback_unlock(TSBACK_master_t *masterp)
{
  _signal_write(&masterp->lock, 0);
}

/*
 * With entry consistency.
 */
INLINE void
tsback_lock_entry(TSBACK_master_t *masterp)
{
  int *lockp = &masterp->lock;

  if (wwt_swap(lockp, 1)) {
    int delay = masterp->delays[wwt_node_number].delay;

    delay = TSBACK_MAX(TSBACK_MIN_DELAY, delay/2);
    wwt_delay(random() % delay);

    while (wwt_swap(lockp, 1))
    {
      delay = TSBACK_MIN(TSBACK_MAX_DELAY, delay*2);
      wwt_delay(random() % delay);
    }

    masterp->delays[wwt_node_number].delay = delay;
  }
}

INLINE void
tsback_unlock_entry(TSBACK_master_t *masterp)
{
  asm volatile ("");
  masterp->lock = 0;
}

#endif _tsback_lock_h
