/*
 * tsbackr_lock.h
 *
 * Copyright (c) 1996 by Jim Goodman
 *
 * ALL RIGHTS RESERVED.
 *
 * This source file is part of the SCI protocol module for the Wisconsin
 * Wind Tunnel, written by Alain K\"agi.
 *
 * ----------------------------------------------------------------------------
 *
 * Test&set with exponential backoff (delay between lock releases).
 */

#ifndef _tsbackr_lock_h
#define _tsbackr_lock_h

#include "assert.h"

/*
 * Declare a lock.
 */
#define TSBACKR_LOCKDEC(lockname)	TSBACKR_master_t lockname

/*
 * Initialize a lock.
 */
#define TSBACKR_LOCKINIT(lockname)					      \
do {									      \
	int __i__;							      \
									      \
	(lockname).delays = (TSBACKR_delay_t *)shm_malloc(wwt_num_nodes	      \
						* sizeof(TSBACKR_delay_t));   \
	assert((lockname).delays != 0);					      \
									      \
	for (__i__ = 0; __i__ < wwt_num_nodes; __i__++) {		      \
		(lockname).delays[__i__].delay = TSBACKR_MIN_DELAY;	      \
	}								      \
	(lockname).lock = 0;						      \
} while (0);

/*
 * Calls.
 */
#define TSBACKR_LOCK(lockname)		tsbackr_lock(&(lockname))
#define TSBACKR_UNLOCK(lockname)	tsbackr_unlock(&(lockname))

#define TSBACKR_LOCK_ENTRY(lockname)	tsbackr_lock_entry(&(lockname))
#define TSBACKR_UNLOCK_ENTRY(lockname)	tsbackr_unlock_entry(&(lockname))

#define TSBACKR_PREFETCH(lockname)
#define TSBACKR_UNPREFETCH(lockname)



/*
 * Details follow.
 */

#define TSBACKR_MIN_DELAY		(64)
#define TSBACKR_MAX_DELAY		(32768 * wwt_num_nodes)
#define TSBACKR_MIN(A,B)		(((A) < (B)) ? (A) : (B))
#define TSBACKR_MAX(A,B)		(((A) > (B)) ? (A) : (B))

#define TSBACKRDEBUG 0
#define TSBACKRPRINTDEBUG if (TSBACKRDEBUG) printf

typedef struct
{
  int delay;
  int pad[15];
} TSBACKR_delay_t;

/*
 * The main structure, one per lock.
 */
typedef struct
{
  TSBACKR_delay_t *delays;
  int pad[15];
  int lock;
} TSBACKR_master_t;

long random();

INLINE void
tsbackr_lock(TSBACKR_master_t *masterp)
{
  volatile int *lockp = &masterp->lock;

  if (wwt_swap_acq(lockp, 1)) {
    int delay = masterp->delays[wwt_node_number].delay;

    while (*lockp) ;

    delay = TSBACKR_MAX(TSBACKR_MIN_DELAY, delay/2);
    wwt_delay(random() % delay);

    while (wwt_swap_acq(lockp, 1))
    {
      while (*lockp) ;

      delay = TSBACKR_MIN(TSBACKR_MAX_DELAY, delay*2);
      wwt_delay(random() % delay);
    }

    masterp->delays[wwt_node_number].delay = delay;
  }
}

INLINE void
tsbackr_unlock(TSBACKR_master_t *masterp)
{
  _signal_write(&masterp->lock, 0);
}

/*
 * With entry consistency.
 */
INLINE void
tsbackr_lock_entry(TSBACKR_master_t *masterp)
{
  volatile int *lockp = &masterp->lock;

  if (wwt_swap(lockp, 1)) {
    int delay = masterp->delays[wwt_node_number].delay;

    while (*lockp) ;

    delay = TSBACKR_MAX(TSBACKR_MIN_DELAY, delay/2);
    wwt_delay(random() % delay);

    while (wwt_swap(lockp, 1))
    {
      while (*lockp) ;

      delay = TSBACKR_MIN(TSBACKR_MAX_DELAY, delay*2);
      wwt_delay(random() % delay);
    }

    masterp->delays[wwt_node_number].delay = delay;
  }
}

INLINE void
tsbackr_unlock_entry(TSBACKR_master_t *masterp)
{
  asm volatile ("");
  masterp->lock = 0;
}

#endif _tsbackr_lock_h
