/*
 * ttsback_lock.h
 *
 * Copyright (c) 1996 by Jim Goodman
 *
 * ALL RIGHTS RESERVED.
 *
 * This source file is part of the SCI protocol module for the Wisconsin
 * Wind Tunnel, written by Alain K\"agi.
 *
 * ----------------------------------------------------------------------------
 *
 * Test&test&set with exponential backoff (delay between references).
 */

#ifndef _ttsback_lock_h
#define _ttsback_lock_h

#include "assert.h"

/*
 * Declare a lock.
 */
#define TTSBACK_LOCKDEC(lockname)	TTSBACK_master_t lockname

/*
 * Initialize a lock.
 */
#define TTSBACK_LOCKINIT(lockname)					      \
do {									      \
	int __i__;							      \
									      \
	(lockname).delays = (TTSBACK_delay_t *)shm_malloc(wwt_num_nodes	      \
						* sizeof(TTSBACK_delay_t));   \
	assert((lockname).delays != 0);					      \
									      \
	for (__i__ = 0; __i__ < wwt_num_nodes; __i__++) {		      \
		(lockname).delays[__i__].delay = TTSBACK_MIN_DELAY;	      \
	}								      \
	(lockname).lock = 0;						      \
} while (0);

/*
 * Calls.
 */
#define TTSBACK_LOCK(lockname)		ttsback_lock(&(lockname))
#define TTSBACK_UNLOCK(lockname)	ttsback_unlock(&(lockname))

#define TTSBACK_LOCK_ENTRY(lockname)	ttsback_lock_entry(&(lockname))
#define TTSBACK_UNLOCK_ENTRY(lockname)	ttsback_unlock_entry(&(lockname))

#define TTSBACK_PREFETCH(lockname)
#define TTSBACK_UNPREFETCH(lockname)



/*
 * Details follow.
 */

#define TTSBACK_MIN_DELAY		(2)
#define TTSBACK_MAX_DELAY		(1024 * wwt_num_nodes)
#define TTSBACK_MIN(A,B)		(((A) < (B)) ? (A) : (B))
#define TTSBACK_MAX(A,B)		(((A) > (B)) ? (A) : (B))

#define TTSBACKDEBUG 0
#define TTSBACKPRINTDEBUG if (TTSBACKDEBUG) printf

typedef struct
{
  int delay;
  int pad[15];
} TTSBACK_delay_t;

/*
 * The main structure, one per lock.
 */
typedef struct
{
  TTSBACK_delay_t *delays;
  int pad[15];
  int lock;
} TTSBACK_master_t;

long random();

INLINE void
ttsback_lock(TTSBACK_master_t *masterp)
{
  volatile int *lockp = &masterp->lock;

  if (*lockp || wwt_swap_acq(lockp, 1)) {
    int delay = masterp->delays[wwt_node_number].delay;

    delay = TTSBACK_MAX(TTSBACK_MIN_DELAY, delay/2);
    wwt_delay(random() % delay);

    while (*lockp || wwt_swap_acq(lockp, 1))
    {
      delay = TTSBACK_MIN(TTSBACK_MAX_DELAY, delay*2);
      wwt_delay(random() % delay);
    }

    masterp->delays[wwt_node_number].delay = delay;
  }
}

INLINE void
ttsback_unlock(TTSBACK_master_t *masterp)
{
  _signal_write(&masterp->lock, 0);
}

/*
 * With entry consistency.
 */
INLINE void
ttsback_lock_entry(TTSBACK_master_t *masterp)
{
  volatile int *lockp = &masterp->lock;

  if (*lockp || wwt_swap(lockp, 1)) {
    int delay = masterp->delays[wwt_node_number].delay;

    delay = TTSBACK_MAX(TTSBACK_MIN_DELAY, delay/2);
    wwt_delay(random() % delay);

    while (*lockp || wwt_swap(lockp, 1))
    {
      delay = TTSBACK_MIN(TTSBACK_MAX_DELAY, delay*2);
      wwt_delay(random() % delay);
    }

    masterp->delays[wwt_node_number].delay = delay;
  }
}

INLINE void
ttsback_unlock_entry(TTSBACK_master_t *masterp)
{
  asm volatile ("");
  masterp->lock = 0;
}

#endif _ttsback_lock_h

#if 0
/*
 * Markup version w/ monitoring code.
 */
INLINE void
ttsback_lock(TTSBACK_master_t *masterp)
{
  int *lockp = &masterp->lock;

  n_synch++;

  if (*lockp || wwt_swap_acq(lockp, 1)) {
    int delay = masterp->delays[wwt_node_number].delay;

    delay = TTSBACK_MAX(TTSBACK_MIN_DELAY, delay/2);

    n_spin++;
    logdelay = TTSBACK_MAX(TTSBACK_LOG_MIN_DELAY, logdelay-1);
    delays[logdelay] += 1;
    TTSBACKPRINTDEBUG("pe%d: failed, initial delay = %d\n",
		      wwt_node_number, delay);

    wwt_delay(random() % delay);

    while (*lockp || wwt_swap_acq(lockp, 1))
    {
      delay = TTSBACK_MIN(TTSBACK_MAX_DELAY, delay*2);

      n_spin++;
      logdelay = TTSBACK_MIN(TTSBACK_LOG_MAX_DELAY, logdelay+1);
      delays[logdelay] += 1;
      TTSBACKPRINTDEBUG("pe%d: failed, delay = %d\n",
			wwt_node_number, delay);

      wwt_delay(random() % delay);
    }

    masterp->delays[wwt_node_number].delay = delay;
  }

  TTSBACKPRINTDEBUG("pe%d: success\n", wwt_node_number);
}

/*
 * Globals
 */
#define TTSBACK_LOG_MIN_DELAY (1)
#define TTSBACK_LOG_MAX_DELAY (17)
int n_synch;
int n_spin;
int delays[256];
int logdelay;

/*
 * The following was in the LOCKINIT macro
 */
									      \
	logdelay = TTSBACK_LOG_MIN_DELAY;				      \
	n_synch = 0;							      \
	n_spin = 0;							      \
	for (__i__ = 0; __i__ < 256; __i__++) {	      \
		delays[__i__] = 0;					      \
	}								      \

/*
 * The old code.
 */
INLINE void
old_ttsback_lock(TTSBACK_master_t *masterp)
{
  volatile int *lockp = &masterp->lock;
  int delay = masterp->delays[wwt_node_number].delay;

  n_synch++;

  if (*lockp || wwt_swap_acq(lockp, 1)) {
    delay = delay/2;    /* first delay = 1/2 of last delay */
    if (delay == 0) {
      delay = 1;
    }
    logdelay = TTSBACK_MAX(0, logdelay-1);

    do {
      n_spin++;
      delays[logdelay] += 1;

      wwt_delay(random() % (delay * 800));
      delay *= 2;
      if (delay > 32) {
        delay = 32;
      }
      logdelay = TTSBACK_MIN(5, logdelay+1);
    }
    while (*lockp || wwt_swap_acq(lockp, 1));
    masterp->delays[wwt_node_number].delay = delay;
  }
}

INLINE void
old_ttsback_unlock(TTSBACK_master_t *masterp)
{
  _signal_write(&masterp->lock, 0);
}

#endif
