/*
 * ttsbackr_lock.h
 *
 * Copyright (c) 1996 by Jim Goodman
 *
 * ALL RIGHTS RESERVED.
 *
 * This source file is part of the SCI protocol module for the Wisconsin
 * Wind Tunnel, written by Alain K\"agi.
 *
 * ----------------------------------------------------------------------------
 *
 * Test&test&set with exponential backoff (delay between lock releases).
 */

#ifndef _ttsbackr_lock_h
#define _ttsbackr_lock_h

#include "assert.h"

/*
 * Declare a lock.
 */
#define TTSBACKR_LOCKDEC(lockname)	TTSBACKR_master_t lockname

/*
 * Initialize a lock.
 */
#define TTSBACKR_LOCKINIT(lockname)					      \
do {									      \
	int __i__;							      \
									      \
	(lockname).delays = (TTSBACKR_delay_t *)shm_malloc(wwt_num_nodes      \
						* sizeof(TTSBACKR_delay_t));  \
	assert((lockname).delays != 0);					      \
									      \
	for (__i__ = 0; __i__ < wwt_num_nodes; __i__++) {		      \
		(lockname).delays[__i__].delay = TTSBACKR_MIN_DELAY;	      \
	}								      \
	(lockname).lock = 0;						      \
} while (0);

/*
 * Calls.
 */
#define TTSBACKR_LOCK(lockname)		ttsbackr_lock(&(lockname))
#define TTSBACKR_UNLOCK(lockname)	ttsbackr_unlock(&(lockname))

#define TTSBACKR_LOCK_ENTRY(lockname)	ttsbackr_lock_entry(&(lockname))
#define TTSBACKR_UNLOCK_ENTRY(lockname)	ttsbackr_unlock_entry(&(lockname))

#define TTSBACKR_PREFETCH(lockname)
#define TTSBACKR_UNPREFETCH(lockname)



/*
 * Details follow.
 */

#define TTSBACKR_MIN_DELAY		(64)
#define TTSBACKR_MAX_DELAY		(32768 * wwt_num_nodes)
#define TTSBACKR_MIN(A,B)		(((A) < (B)) ? (A) : (B))
#define TTSBACKR_MAX(A,B)		(((A) > (B)) ? (A) : (B))

#define TTSBACKRDEBUG 0
#define TTSBACKRPRINTDEBUG if (TTSBACKRDEBUG) printf

typedef struct
{
  int delay;
  int pad[15];
} TTSBACKR_delay_t;

/*
 * The main structure, one per lock.
 */
typedef struct
{
  TTSBACKR_delay_t *delays;
  int pad[15];
  int lock;
} TTSBACKR_master_t;

long random();

INLINE void
ttsbackr_lock(TTSBACKR_master_t *masterp)
{
  volatile int *lockp = &masterp->lock;

  if (*lockp || wwt_swap_acq(lockp, 1)) {
    int delay = masterp->delays[wwt_node_number].delay;

    while (*lockp) ;

    delay = TTSBACKR_MAX(TTSBACKR_MIN_DELAY, delay/2);
    wwt_delay(random() % delay);

    while (*lockp || wwt_swap_acq(lockp, 1))
    {
      while (*lockp) ;

      delay = TTSBACKR_MIN(TTSBACKR_MAX_DELAY, delay*2);
      wwt_delay(random() % delay);
    }

    masterp->delays[wwt_node_number].delay = delay;
  }
}

INLINE void
ttsbackr_unlock(TTSBACKR_master_t *masterp)
{
  _signal_write(&masterp->lock, 0);
}

/*
 * With entry consistency.
 */
INLINE void
ttsbackr_lock_entry(TTSBACKR_master_t *masterp)
{
  volatile int *lockp = &masterp->lock;

  if (*lockp || wwt_swap(lockp, 1)) {
    int delay = masterp->delays[wwt_node_number].delay;

    while (*lockp) ;

    delay = TTSBACKR_MAX(TTSBACKR_MIN_DELAY, delay/2);
    wwt_delay(random() % delay);

    while (*lockp || wwt_swap(lockp, 1))
    {
      while (*lockp) ;

      delay = TTSBACKR_MIN(TTSBACKR_MAX_DELAY, delay*2);
      wwt_delay(random() % delay);
    }

    masterp->delays[wwt_node_number].delay = delay;
  }
}

INLINE void
ttsbackr_unlock_entry(TTSBACKR_master_t *masterp)
{
  asm volatile ("");
  masterp->lock = 0;
}

#endif _ttsbackr_lock_h
